@@ -1,5 +1,7 @@
+use std::cmp::Ordering;
+
 use crate::types::Address;
-use rlp::{Encodable, RlpStream};
+use rlp::{Decodable, Encodable, RlpStream};
 use serde::{ser::Error as SerializationError, Deserialize, Deserializer, Serialize, Serializer};
 
 /// ENS name or Ethereum Address. Not RLP encoded/serialized if it's a name
@@ -29,6 +31,25 @@ impl Encodable for NameOrAddress {
     }
 }
 
+impl Decodable for NameOrAddress {
+    fn decode(rlp: &rlp::Rlp) -> Result<Self, rlp::DecoderError> {
+        // An address (H160) is 20 bytes, so let's only accept 20 byte rlp string encodings.
+        if !rlp.is_data() {
+            return Err(rlp::DecoderError::RlpExpectedToBeData)
+        }
+
+        // the data needs to be 20 bytes long
+        match 20.cmp(&rlp.size()) {
+            Ordering::Less => Err(rlp::DecoderError::RlpIsTooShort),
+            Ordering::Greater => Err(rlp::DecoderError::RlpIsTooBig),
+            Ordering::Equal => {
+                let rlp_data = rlp.data()?;
+                Ok(NameOrAddress::Address(Address::from_slice(rlp_data)))
+            }
+        }
+    }
+}
+
 impl From<&str> for NameOrAddress {
     fn from(s: &str) -> Self {
         NameOrAddress::Name(s.to_owned())
@@ -71,6 +92,8 @@ impl<'de> Deserialize<'de> for NameOrAddress {
 
 #[cfg(test)]
 mod tests {
+    use rlp::Rlp;
+
     use super::*;
 
     #[test]
@@ -82,7 +105,7 @@ mod tests {
         assert!(rlp.is_empty());
 
         let mut rlp = RlpStream::new();
-        (&name).rlp_append(&mut rlp);
+        name.rlp_append(&mut rlp);
         assert!(rlp.is_empty());
     }
 
@@ -99,10 +122,24 @@ mod tests {
         assert_eq!(rlp.as_raw(), expected.as_raw());
 
         let mut rlp = RlpStream::new();
-        (&union).rlp_append(&mut rlp);
+        union.rlp_append(&mut rlp);
         assert_eq!(rlp.as_raw(), expected.as_raw());
     }
 
+    #[test]
+    fn rlp_address_deserialized() {
+        let addr = "3dd6f334b732d23b51dfbee2070b40bbd1a97a8f".parse().unwrap();
+        let expected = NameOrAddress::Address(addr);
+
+        let mut rlp = RlpStream::new();
+        rlp.append(&addr);
+        let rlp_bytes = &rlp.out().freeze()[..];
+        let data = Rlp::new(rlp_bytes);
+        let name = NameOrAddress::decode(&data).unwrap();
+
+        assert_eq!(name, expected);
+    }
+
     #[test]
     fn serde_name_not_serialized() {
         let name = NameOrAddress::Name("ens.eth".to_string());
